==> create-mashup-app-2.0.2.tgz: package/src/plugins/mashupRollupPlugin.js <==
const {walk} = require('estree-walker');
const MagicString = require('magic-string');
const Concat = require('concat-with-sourcemaps');

const PLUGIN_NAME = 'mashup-entry-plugin';

/**
 * Installs a dynamic module loader for this mashup.
 * It will create a new global with name `mashupImportGlobalName`
 * and initialize it with a function that loads another chunk
 * and returns a promise for that chunk. It will resolve immediately
 * if the chunk is already loaded on the page.
 */
function getMashupModuleLoader({
  mashupDependenciesGlobalName,
  mashupChunksGlobalName,
  mashupImportGlobalName,
  basePath
}) {
  return `
window.${mashupChunksGlobalName} = window.${mashupChunksGlobalName} || {};
window.${mashupDependenciesGlobalName} = {};
window.${mashupImportGlobalName} = (function(moduleId, moduleReady) {
    var pending = this;
    if (moduleReady) {
        if (pending[moduleId]) {
            pending[moduleId].resolve(window.${mashupChunksGlobalName}[moduleId]);
            delete pending[moduleId];
        }
        return;
    }
    if (window.${mashupChunksGlobalName}[moduleId]) {
        return Promise.resolve(window.${mashupChunksGlobalName}[moduleId]);
    }
    if (!pending[moduleId]) {
        pending[moduleId] = {};
        pending[moduleId].promise = new Promise(function(resolve){
            pending[moduleId].resolve = resolve;
            var scriptElement = document.createElement('script');
            scriptElement.setAttribute('src', '${basePath}' + '/' + moduleId);
            document.body.appendChild(scriptElement);
        });
    }
    return pending[moduleId].promise;
}).bind({});
`;
}

function getMashupInitializer({
  dependencies,
  target,
  mashupModuleName,
  mashupChunksGlobalName,
  mashupDependenciesGlobalName,
  mashupGlobalName,
  entryChunkName
}) {
  return `
window.${mashupGlobalName} = window.${mashupChunksGlobalName}['./${entryChunkName}'];
tau.mashups
${dependencies
    .map(item => `.addDependency('${item.mashupDependency || item}')`)
    .join('\n')}
.${target === 'mashup' ? 'addMashup(' : `addModule('${mashupModuleName}',`}
function() {
    ${dependencies
      .map(
        (item, index) =>
          `window.${mashupDependenciesGlobalName}['${item.provideModule ||
            item.mashupDependency ||
            item}'] = arguments[${index}];`
      )
      .join('\n')}
    ${
      target === 'mashup'
        ? `window.${mashupGlobalName}().initialize();`
        : `return window.${mashupGlobalName}();`
    }
});

`;
}

module.exports = function mashup(settings) {
  // options are taken from the object's property to
  // allow them to be overriden in plugins array
  const plugin = {
    name: PLUGIN_NAME,
    settings,
    generateBundle: async function(outputOptions, bundle) {
      // all regular chunks (isEntry: true) are loaded with our miniloader
      // but there are also commons chunks extracted automatically to avoid
      // duplication, and even our entry module may use these
      // so we need to inline them with our entry point chunk
      const {
        entryPoint,
        mashupChunksGlobalName,
        mashupDependenciesGlobalName,
        mashupImportGlobalName,
        basePath,
        target
      } = plugin.settings;
      const commonsChunkNames = Object.keys(bundle).filter(
        name => !bundle[name].isEntry
      );
      const entryPointChunk =
        bundle[
          Object.keys(bundle).find(
            chunkName => bundle[chunkName].modules[entryPoint]
          )
        ];
      const concat = new Concat(true, entryPoint);
      concat.add(
        null,
        getMashupModuleLoader({
          mashupChunksGlobalName,
          mashupDependenciesGlobalName,
          mashupImportGlobalName,
          basePath
        })
      );
      for (const chunkName of commonsChunkNames) {
        const {code, map} = bundle[chunkName];
        concat.add(chunkName, code, JSON.stringify(map));
        delete bundle[chunkName];
      }
      concat.add(
        entryPointChunk.fileName,
        entryPointChunk.code,
        JSON.stringify(entryPointChunk.map)
      );

      entryPointChunk.code = concat.content.toString();
      entryPointChunk.map = JSON.parse(concat.sourceMap);

      if (target === 'mashup') {
        bundle['manifest.cfg'] = {code: 'Placeholders:restui_board'};
      }
    },
    renderChunk: async function(sourceCode, options) {
      const {
        entryPoint,
        basePath,
        dependencies,
        target,
        mashupModuleName,
        mashupGlobalName,
        mashupImportGlobalName,
        mashupChunksGlobalName,
        mashupDependenciesGlobalName
      } = plugin.settings;

      // our global names are window.<something>_chunks['<chunk id>']
      // but iife plugin treats those as variable names and prepends var
      // we insert spaces instead of var here to preserve symbol position
      // so that we don't have to merge sourcemaps
      sourceCode = sourceCode.split('var window.').join('    window.');

      const code = new MagicString(sourceCode);
      const ast = this.parse(code.toString(), {
        // we also need to remove all comments because require('<something>') in comments
        // causes TP SystemJS loader to treat it as external module to be loaded
        onComment(isBlock, text, start, end) {
          code.overwrite(start, end, '');
        }
      });
      // if chunk contains entry point module, it is an entry point chunk
      if (options.modules[entryPoint]) {
        const [initStatement] = ast.body;
        // window.<something> = <right>, RHS is IIFE for entry point
        // we need to replace it with regular function, because it can only be called after
        // all mashup dependencies are assigned properly
        const assignmentExpression = initStatement.expression;
        const iifeCallExpression = assignmentExpression.right;
        const functionExpression = iifeCallExpression.callee;

        // basically replace function call with callee, so we get a function and not IIFE
        code.overwrite(iifeCallExpression.start, functionExpression.start, '');
        code.overwrite(functionExpression.end, iifeCallExpression.end, '');

        code.appendRight(
          ast.body[ast.body.length - 1].end,
          getMashupInitializer({
            dependencies,
            target,
            mashupModuleName,
            mashupDependenciesGlobalName,
            mashupChunksGlobalName,
            mashupGlobalName,
            entryChunkName: options.fileName
          })
        );
      } else {
        // otherwise it's a dynamically-loaded chunk that needs to
        // notify the awaiting module that it's loaded
        code.appendRight(
          ast.body[ast.body.length - 1].end,
          `window.${mashupImportGlobalName}('./${options.fileName}', true);`
        );
        // if there are no dynamically-loaded chunks, basePath is not required
        // but if we are here, some of the chunks were dynamic and we need to throw
        // on missing basePath

        if (!basePath) {
          throw new Error(
            `basePath not specified, but ${
              options.fileName
            } is loaded with dynamic import().
Code splitting requires basePath from which all of your output chunks are avaliable.`
          );
        }
      }

      // regardless of module type, we need to replace dynamic imports
      // with our own import function, and naive string replace could
      // malform strings/comments/etc.
      walk(ast, {
        enter(node, parent) {
          // Import node as callee of CallExpression is dynamic import call
          if (node.type === 'Import' && parent.type === 'CallExpression') {
            // node.start and node.end correspond to 6 characters - 'import'
            // we replace these with a regular function call - <something>_import
            code.overwrite(node.start, node.end, mashupImportGlobalName);
          }
        }
      });

      return {
        code: code.toString(),
        map: code.generateMap({hires: true})
      };
    }
  };

  return plugin;
};

module.exports.pluginName = PLUGIN_NAME;

==> create-mashup-app-2.4.0.tgz: package/src/plugins/mashupRollupPlugin.js <==
const {walk} = require('estree-walker');
const MagicString = require('magic-string');
const Concat = require('concat-with-sourcemaps');

const PLUGIN_NAME = 'mashup-entry-plugin';

/**
 * Installs a dynamic module loader for this mashup.
 * It will create a new global with name `mashupImportGlobalName`
 * and initialize it with a function that loads another chunk
 * and returns a promise for that chunk. It will resolve immediately
 * if the chunk is already loaded on the page.
 */
function getMashupModuleLoader({
  mashupDependenciesGlobalName,
  mashupChunksGlobalName,
  mashupImportGlobalName,
  basePath
}) {
  return `
window.${mashupChunksGlobalName} = window.${mashupChunksGlobalName} || {};
window.${mashupDependenciesGlobalName} = {};
window.${mashupImportGlobalName} = (function(moduleId, moduleReady) {
    var pending = this;
    if (moduleReady) {
        if (pending[moduleId]) {
            pending[moduleId].resolve(window.${mashupChunksGlobalName}[moduleId]);
            delete pending[moduleId];
        }
        return;
    }
    if (window.${mashupChunksGlobalName}[moduleId]) {
        return Promise.resolve(window.${mashupChunksGlobalName}[moduleId]);
    }
    if (!pending[moduleId]) {
        pending[moduleId] = {};
        pending[moduleId].promise = new Promise(function(resolve){
            pending[moduleId].resolve = resolve;
            var scriptElement = document.createElement('script');
            scriptElement.setAttribute('src', '${basePath}' + '/' + moduleId);
            document.body.appendChild(scriptElement);
        });
    }
    return pending[moduleId].promise;
}).bind({});
`;
}

function getMashupInitializer({
  dependencies,
  target,
  mashupModuleName,
  mashupChunksGlobalName,
  mashupDependenciesGlobalName,
  mashupGlobalName,
  entryChunkName
}) {
  return `
window.${mashupGlobalName} = window.${mashupChunksGlobalName}['./${entryChunkName}'];
tau.mashups
${dependencies
    .map(item => `.addDependency('${item.mashupDependency || item}')`)
    .join('\n')}
.${target === 'mashup' ? 'addMashup(' : `addModule('${mashupModuleName}',`}
function() {
    ${dependencies
      .map(
        (item, index) =>
          `window.${mashupDependenciesGlobalName}['${item.provideModule ||
            item.mashupDependency ||
            item}'] = arguments[${index}];`
      )
      .join('\n')}
    ${
      target === 'mashup'
        ? `window.${mashupGlobalName}().initialize();`
        : `return window.${mashupGlobalName}();`
    }
});

`;
}

module.exports = function mashup(settings) {
  // options are taken from the object's property to
  // allow them to be overriden in plugins array
  const plugin = {
    name: PLUGIN_NAME,
    settings,
    generateBundle: async function(outputOptions, bundle) {
      // all regular chunks (isEntry: true) are loaded with our miniloader
      // but there are also commons chunks extracted automatically to avoid
      // duplication, and even our entry module may use these
      // so we need to inline them with our entry point chunk
      const {
        entryPoint,
        mashupChunksGlobalName,
        mashupDependenciesGlobalName,
        mashupImportGlobalName,
        basePath,
        target
      } = plugin.settings;
      const commonsChunkNames = Object.keys(bundle).filter(
        name => !bundle[name].isEntry
      );
      const entryPointChunk =
        bundle[
          Object.keys(bundle).find(
            chunkName => bundle[chunkName].modules[entryPoint]
          )
        ];
      const concat = new Concat(true, entryPoint);
      concat.add(
        null,
        getMashupModuleLoader({
          mashupChunksGlobalName,
          mashupDependenciesGlobalName,
          mashupImportGlobalName,
          basePath
        })
      );
      for (const chunkName of commonsChunkNames) {
        const {code, map} = bundle[chunkName];
        concat.add(chunkName, code, JSON.stringify(map));
        delete bundle[chunkName];
      }
      concat.add(
        entryPointChunk.fileName,
        entryPointChunk.code,
        JSON.stringify(entryPointChunk.map)
      );

      entryPointChunk.code = concat.content.toString();
      if (outputOptions.sourcemap) {
        entryPointChunk.map.mappings = JSON.parse(concat.sourceMap).mappings;
      }

      if (target === 'mashup') {
        bundle['manifest.cfg'] = {code: 'Placeholders:restui_board'};
      }
    },
    renderChunk: async function(sourceCode, options) {
      const {
        entryPoint,
        basePath,
        dependencies,
        target,
        mashupModuleName,
        mashupGlobalName,
        mashupImportGlobalName,
        mashupChunksGlobalName,
        mashupDependenciesGlobalName
      } = plugin.settings;

      // our global names are window.<something>_chunks['<chunk id>']
      // but iife plugin treats those as variable names and prepends var
      // we insert spaces instead of var here to preserve symbol position
      // so that we don't have to merge sourcemaps
      sourceCode = sourceCode.split('var window.').join('    window.');

      const code = new MagicString(sourceCode);
      const ast = this.parse(code.toString(), {
        // we also need to remove all comments because require('<something>') in comments
        // causes TP SystemJS loader to treat it as external module to be loaded
        onComment(isBlock, text, start, end) {
          code.overwrite(start, end, '');
        }
      });
      // if chunk contains entry point module, it is an entry point chunk
      if (options.modules[entryPoint]) {
        const [initStatement] = ast.body;
        // window.<something> = <right>, RHS is IIFE for entry point
        // we need to replace it with regular function, because it can only be called after
        // all mashup dependencies are assigned properly
        const assignmentExpression = initStatement.expression;
        if (assignmentExpression.type !== 'AssignmentExpression') {
          throw new Error(
            `Mashup entry point module ${entryPoint} doesn't export anything. It must have a named export function 'initialize'. Any toplevel code using dependencies before 'initialize' call is invalid.`
          );
        }
        const iifeCallExpression = assignmentExpression.right;
        const functionExpression = iifeCallExpression.callee;

        // basically replace function call with callee, so we get a function and not IIFE
        code.overwrite(iifeCallExpression.start, functionExpression.start, '');
        code.overwrite(functionExpression.end, iifeCallExpression.end, '');

        code.appendRight(
          ast.body[ast.body.length - 1].end,
          getMashupInitializer({
            dependencies,
            target,
            mashupModuleName,
            mashupDependenciesGlobalName,
            mashupChunksGlobalName,
            mashupGlobalName,
            entryChunkName: options.fileName
          })
        );
      } else {
        // otherwise it's a dynamically-loaded chunk that needs to
        // notify the awaiting module that it's loaded
        code.appendRight(
          ast.body[ast.body.length - 1].end,
          `window.${mashupImportGlobalName}('./${options.fileName}', true);`
        );
        // if there are no dynamically-loaded chunks, basePath is not required
        // but if we are here, some of the chunks were dynamic and we need to throw
        // on missing basePath

        if (!basePath) {
          throw new Error(
            `basePath not specified, but ${
              options.fileName
            } is loaded with dynamic import().
Code splitting requires basePath from which all of your output chunks are avaliable.`
          );
        }
      }

      // regardless of module type, we need to replace dynamic imports
      // with our own import function, and naive string replace could
      // malform strings/comments/etc.
      walk(ast, {
        enter(node, parent) {
          // Import node as callee of CallExpression is dynamic import call
          if (node.type === 'Import' && parent.type === 'CallExpression') {
            // node.start and node.end correspond to 6 characters - 'import'
            // we replace these with a regular function call - <something>_import
            code.overwrite(node.start, node.end, mashupImportGlobalName);
          }
        }
      });

      return {
        code: code.toString(),
        map: code.generateMap({hires: true})
      };
    }
  };

  return plugin;
};

module.exports.pluginName = PLUGIN_NAME;
